# Copyright (C) 2020-2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

# -----------------------------------------------------------
# Primary author: Hongyan Chang <hongyan.chang@intel.com>
# Co-authored-by: Anindya S. Paul <anindya.s.paul@intel.com>
# Co-authored-by: Brandon Edwards <brandon.edwards@intel.com>
# ------------------------------------------------------------

from copy import deepcopy
import torch.nn as nn
import torch.optim as optim
import torch
import numpy as np
from openfl.experimental.workflow.interface import FLSpec, Aggregator, Collaborator
from openfl.experimental.workflow.runtime import LocalRuntime
from openfl.experimental.workflow.placement import aggregator, collaborator
import torchvision.transforms as transforms
import pickle
from pathlib import Path

from privacy_meter.model import PytorchModelTensor
import copy
from auditor import (
    PopulationAuditor,
    plot_auc_history,
    plot_tpr_history,
    plot_roc_history,
    PM_report,
)

import time
import os
import argparse
from cifar10_loader import CIFAR10
import warnings

warnings.filterwarnings("ignore")

batch_size_train = 32
batch_size_test = 1000
learning_rate = 0.005
momentum = 0.9
log_interval = 10

# set the random seed for repeatable results
random_seed = 10
torch.manual_seed(random_seed)


class Net(nn.Module):
    def __init__(self, num_classes=10):
        super(Net, self).__init__()
        self.features = nn.Sequential(
            nn.Conv2d(3, 64, kernel_size=3, stride=2, padding=1),
            nn.ReLU(inplace=True),
            nn.MaxPool2d(kernel_size=2),
            nn.Conv2d(64, 192, kernel_size=3, padding=1),
            nn.ReLU(inplace=True),
            nn.MaxPool2d(kernel_size=2),
            nn.Conv2d(192, 384, kernel_size=3, padding=1),
            nn.ReLU(inplace=True),
            nn.Conv2d(384, 256, kernel_size=3, padding=1),
            nn.ReLU(inplace=True),
            nn.Conv2d(256, 256, kernel_size=3, padding=1),
            nn.ReLU(inplace=True),
            nn.MaxPool2d(kernel_size=2),
        )
        self.classifier = nn.Sequential(
            nn.Dropout(),
            nn.Linear(256 * 2 * 2, 4096),
            nn.ReLU(inplace=True),
            nn.Dropout(),
            nn.Linear(4096, 4096),
            nn.ReLU(inplace=True),
            nn.Linear(4096, num_classes),
        )

    def forward(self, x):
        x = self.features(x)
        x = x.view(x.size(0), 256 * 2 * 2)
        x = self.classifier(x)
        return x


def default_optimizer(model, optimizer_type=None, optimizer_like=None):
    """
    Return a new optimizer based on the optimizer_type or the optimizer template

    Args:
        model:   NN model architected from nn.module class
        optimizer_type: "SGD" or "Adam"
        optimizer_like: "torch.optim.SGD" or "torch.optim.Adam" optimizer
    """
    if optimizer_type == "SGD" or isinstance(optimizer_like, optim.SGD):
        return optim.SGD(model.parameters(), lr=learning_rate, momentum=momentum)
    elif optimizer_type == "Adam" or isinstance(optimizer_like, optim.Adam):
        return optim.Adam(model.parameters())


def FedAvg(models):  # NOQA: N802
    """
    Return a Federated average model based on Fedavg algorithm: H. B. Mcmahan,
    E. Moore, D. Ramage, S. Hampson, and B. A. Y.Arcas,
    “Communication-efficient learning of deep networks from decentralized data,” 2017.

    Args:
        models: Python list of locally trained models by each collaborator
    """
    new_model = models[0]
    if len(models) > 1:
        state_dicts = [model.state_dict() for model in models]
        state_dict = new_model.state_dict()
        for key in models[1].state_dict():
            state_dict[key] = np.sum(
                [state[key] for state in state_dicts], axis=0
            ) / len(models)
        # Convert numpy arrays within the state dictionary to PyTorch tensors
        state_dict_tensors = {key: torch.tensor(value, dtype=torch.float32).cpu() for key, value in state_dict.items()}

        # Load the converted state dictionary into the model
        new_model.load_state_dict(state_dict_tensors)
    return new_model


def inference(network, test_loader, device):
    network.eval()
    network.to(device)
    test_loss = 0
    correct = 0
    with torch.no_grad():
        for data, target in test_loader:
            data = data.to(device)
            target = target.to(device)
            output = network(data)
            criterion = nn.CrossEntropyLoss()
            test_loss += criterion(output, target).item()
            pred = output.data.max(1, keepdim=True)[1]
            correct += pred.eq(target.data.view_as(pred)).sum()
    test_loss /= len(test_loader)
    accuracy = float(correct / len(test_loader.dataset))
    print(
        (
            f"Test set: Avg. loss: {test_loss}, "
            f"Accuracy: {correct}/{len(test_loader.dataset)} ({100.0 * accuracy}%)"
        )
    )
    network.to("cpu")
    return accuracy


def optimizer_to_device(optimizer, device):
    """
    Sending the "torch.optim.Optimizer" object into the specified device
    for model training and inference

    Args:
        optimizer: torch.optim.Optimizer from "default_optimizer" function
        device: CUDA device id or "cpu"
    """
    if optimizer.state_dict()["state"] != {}:
        if isinstance(optimizer, optim.SGD):
            for param in optimizer.param_groups[0]["params"]:
                param.data = param.data.to(device)
                if param.grad is not None:
                    param.grad = param.grad.to(device)
        elif isinstance(optimizer, optim.Adam):
            for state in optimizer.state.values():
                for k, v in state.items():
                    if isinstance(v, torch.Tensor):
                        state[k] = v.to(device)
    else:
        raise (ValueError("No dict keys in optimizer state: please check"))


def load_previous_round_model_and_optimizer_and_perform_testing(
    model, global_model, optimizer, collaborator_name, round_num, device
):
    """
    Load pickle file to retrieve the model and optimizer state dictionary
    from the previous round for each collaborator
    and perform several validation routines with current
    round state dictionaries to test the flow loop.
    Note: this functionality can be enabled through the command line argument
    by setting "--flow_internal_loop_test=True".

    Args:
        model: local collaborator model at the current round
        global_model: Federated averaged model at the aggregator
        optimizer: local collaborator optimizer at the current round
        collaborator_name: name of the collaborator (Type:string)
        round_num: current round (Type:int)
        device: CUDA device id or "cpu"
    """
    print(f"Loading model and optimizer state dict for round {round_num-1}")
    model_prevround = Net()  # instantiate a new model
    model_prevround = model_prevround.to(device)
    optimizer_prevround = default_optimizer(model_prevround, optimizer_like=optimizer)
    if os.path.isfile(
        f"Collaborator_{collaborator_name}_model_config_roundnumber_{round_num-1}.pickle"
    ):
        with open(
            f"Collaborator_{collaborator_name}_model_config_roundnumber_{round_num-1}.pickle",
            "rb",
        ) as f:
            model_prevround_config = pickle.load(f)
            model_prevround.load_state_dict(model_prevround_config["model_state_dict"])
            optimizer_prevround.load_state_dict(
                model_prevround_config["optim_state_dict"]
            )

            for param_tensor in model.state_dict():
                for tensor_1, tensor_2 in zip(
                    model.state_dict()[param_tensor],
                    global_model.state_dict()[param_tensor],
                ):
                    if (
                        torch.equal(tensor_1.to(device), tensor_2.to(device))
                        is not True
                    ):
                        raise (
                            ValueError(
                                (
                                    "local and global model differ: "
                                    f"{collaborator_name} at round {round_num-1}."
                                )
                            )
                        )

                if isinstance(optimizer, optim.SGD):
                    if optimizer.state_dict()["state"] != {}:
                        for param_idx in optimizer.state_dict()["param_groups"][0][
                            "params"
                        ]:
                            for tensor_1, tensor_2 in zip(
                                optimizer.state_dict()["state"][param_idx][
                                    "momentum_buffer"
                                ],
                                optimizer_prevround.state_dict()["state"][param_idx][
                                    "momentum_buffer"
                                ],
                            ):
                                if (
                                    torch.equal(
                                        tensor_1.to(device), tensor_2.to(device)
                                    )
                                    is not True
                                ):
                                    raise (
                                        ValueError(
                                            (
                                                "Momentum buffer data differ: "
                                                f"{collaborator_name} at round {round_num-1}"
                                            )
                                        )
                                    )
                    else:
                        raise (ValueError("Current optimizer state is empty"))

                model_params = [
                    model.state_dict()[param_tensor]
                    for param_tensor in model.state_dict()
                ]
                for idx, param in enumerate(optimizer.param_groups[0]["params"]):
                    for tensor_1, tensor_2 in zip(param.data, model_params[idx]):
                        if (
                            torch.equal(tensor_1.to(device), tensor_2.to(device))
                            is not True
                        ):
                            raise (
                                ValueError(
                                    (
                                        "Model and optimizer do not point "
                                        "to the same params for collaborator: "
                                        f"{collaborator_name} at round {round_num-1}."
                                    )
                                )
                            )

    else:
        raise (ValueError("No such name of pickle file exists"))


def save_current_round_model_and_optimizer_for_next_round_testing(
    model, optimizer, collaborator_name, round_num
):
    """
    Save the model and optimizer state dictionary
    of a collaboartor ("collaborator_name")
    in a given round ("round_num") into a pickle file
    for later retrieving and verifying its correctness.
    This provide the user the ability to verify the fields
    in the model and optimizer state dictionary and
    may provide confidence on the results of privacy auditing.
    Note: this functionality can be enabled through the command line
    argument by setting "--flow_internal_loop_test=True".

    Args:
        model: local collaborator model at the current round
        optimizer: local collaborator optimizer at the current round
        collaborator_name: name of the collaborator (Type:string)
        round_num: current round (Type:int)
    """
    model_config = {
        "model_state_dict": model.state_dict(),
        "optim_state_dict": optimizer.state_dict(),
    }
    with open(
        f"Collaborator_{collaborator_name}_model_config_roundnumber_{round_num}.pickle",
        "wb",
    ) as f:
        pickle.dump(model_config, f)


class FederatedFlow(FLSpec):
    def __init__(
        self,
        model,
        optimizers,
        device="cpu",
        total_rounds=10,
        top_model_accuracy=0,
        flow_internal_loop_test=False,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.model = model
        self.global_model = Net()
        self.optimizers = optimizers
        self.total_rounds = total_rounds
        self.top_model_accuracy = top_model_accuracy
        self.device = device
        self.flow_internal_loop_test = flow_internal_loop_test
        self.round_num = 0  # starting round
        print(20 * "#")
        print(f"Round {self.round_num}...")
        print(20 * "#")

    @aggregator
    def start(self):
        self.start_time = time.time()
        print("Performing initialization for model")
        self.collaborators = self.runtime.collaborators
        self.private = 10
        self.next(
            self.aggregated_model_validation,
            foreach="collaborators",
            exclude=["private"],
        )

    @collaborator
    def aggregated_model_validation(self):
        print(
            (
                "Performing aggregated model validation for collaborator: "
                f"{self.input} in round {self.round_num}"
            )
        )
        self.agg_validation_score = inference(self.model, self.test_loader, self.device)
        print(f"{self.input} value of {self.agg_validation_score}")
        self.collaborator_name = self.input
        self.next(self.train)

    @collaborator
    def train(self):
        print(20 * "#")
        print(
            f"Performing model training for collaborator {self.input} in round {self.round_num}"
        )

        self.model.to(self.device)
        self.optimizer = default_optimizer(
            self.model, optimizer_like=self.optimizers[self.input]
        )

        if self.round_num > 0:
            self.optimizer.load_state_dict(
                deepcopy(self.optimizers[self.input].state_dict())
            )
            optimizer_to_device(optimizer=self.optimizer, device=self.device)

            if self.flow_internal_loop_test:
                load_previous_round_model_and_optimizer_and_perform_testing(
                    self.model,
                    self.global_model,
                    self.optimizer,
                    self.collaborator_name,
                    self.round_num,
                    self.device,
                )

        self.model.train()
        train_losses = []
        for batch_idx, (data, target) in enumerate(self.train_loader):
            data = data.to(self.device)
            target = target.to(self.device)
            self.optimizer.zero_grad()
            output = self.model(data)
            criterion = nn.CrossEntropyLoss()
            loss = criterion(output, target).to(self.device)
            loss.backward()
            self.optimizer.step()
            if batch_idx % log_interval == 0:
                train_losses.append(loss.item())

        self.loss = np.mean(train_losses)
        self.training_completed = True

        if self.flow_internal_loop_test:
            save_current_round_model_and_optimizer_for_next_round_testing(
                self.model, self.optimizer, self.collaborator_name, self.round_num
            )

        self.model.to("cpu")
        tmp_opt = deepcopy(self.optimizers[self.input])
        tmp_opt.load_state_dict(self.optimizer.state_dict())
        self.optimizer = tmp_opt
        torch.cuda.empty_cache()
        self.next(self.local_model_validation)

    @collaborator
    def local_model_validation(self):
        print(
            (
                "Performing local model validation for collaborator: "
                f"{self.input} in round {self.round_num}"
            )
        )
        print(self.device)
        start_time = time.time()

        print("Test dataset performance")
        self.local_validation_score = inference(
            self.model, self.test_loader, self.device
        )
        print("Train dataset performance")
        self.local_validation_score_train = inference(
            self.model, self.train_loader, self.device
        )

        print(
            (
                "Doing local model validation for collaborator: "
                f"{self.input}: {self.local_validation_score}"
            )
        )
        print(f"local validation time cost {(time.time() - start_time)}")

        if (
            self.round_num == 0
            or self.round_num % self.local_pm_info.interval == 0
            or self.round_num == self.total_rounds
        ):
            print("Performing Auditing")
            self.next(self.audit)
        else:
            self.next(self.join, exclude=["training_completed"])

    @collaborator
    def audit(self):
        print(
            (
                "Performing local and global model auditing for collaborator: "
                f"{self.input} in round {self.round_num}"
            )
        )
        begin_time = time.time()

        datasets = {
            "train": self.train_dataset,
            "test": self.test_dataset,
            "audit": self.population_dataset,
        }

        start_time = time.time()
        # batch_size for the PytorchModelTensor indicates batch size for computing the signals.
        # for computing loss and logits, it can be large, e.g., 1000.
        # for computing the signal_norm, it should be around 25.
        # Otherwise, one may get OOM depending on the GPU memory.

        target_model = PytorchModelTensor(
            copy.deepcopy(self.model), nn.CrossEntropyLoss(), self.device
        )
        self.local_pm_info = PopulationAuditor(
            target_model, datasets, self.local_pm_info
        )
        target_model.model_obj.to("cpu")
        self.local_pm_info.update_history("round", self.round_num)

        print(f"population attack for the local model uses {time.time() - start_time}")

        start_time = time.time()
        target_model = PytorchModelTensor(
            copy.deepcopy(self.global_model), nn.CrossEntropyLoss(), self.device
        )
        self.global_pm_info = PopulationAuditor(
            target_model, datasets, self.global_pm_info
        )
        self.global_pm_info.update_history("round", self.round_num)
        target_model.model_obj.to("cpu")
        print(f"population attack for the global model uses {time.time() - start_time}")

        start_time = time.time()

        history_dict = {
            "PM Result (Local)": self.local_pm_info,
            "PM Result (Global)": self.global_pm_info,
        }

        # # generate the plot for the privacy loss
        plot_tpr_history(history_dict, self.input, self.local_pm_info.fpr_tolerance)
        plot_auc_history(history_dict, self.input)
        plot_roc_history(history_dict, self.input)

        # save the privacy report
        saving_path = f"{self.local_pm_info.log_dir}/{self.input}.pkl"
        Path(self.local_pm_info.log_dir).mkdir(parents=True, exist_ok=True)
        with open(saving_path, "wb") as handle:
            pickle.dump(history_dict, handle, protocol=pickle.HIGHEST_PROTOCOL)
        print(f"auditing time: {time.time() - begin_time}")

        # Clean up state before transitioning to collaborator
        delattr(self, "train_dataset")
        delattr(self, "train_loader")
        delattr(self, "test_dataset")
        delattr(self, "test_loader")
        delattr(self, "population_dataset")
        self.next(self.join, exclude=["training_completed"])

    @aggregator
    def join(self, inputs):
        self.average_loss = sum(input.loss for input in inputs) / len(inputs)
        self.aggregated_model_accuracy = sum(
            input.agg_validation_score for input in inputs
        ) / len(inputs)
        self.local_model_accuracy = sum(
            input.local_validation_score for input in inputs
        ) / len(inputs)
        print(
            f"Average aggregated model validation values = {self.aggregated_model_accuracy}"
        )
        print(f"Average training loss = {self.average_loss}")
        print(f"Average local model validation values = {self.local_model_accuracy}")

        self.model = FedAvg([input.model.cpu() for input in inputs])
        self.global_model.load_state_dict(deepcopy(self.model.state_dict()))
        self.optimizers.update(
            {input.collaborator_name: input.optimizer for input in inputs}
        )

        del inputs
        self.next(self.check_round_completion)

    @aggregator
    def check_round_completion(self):
        if self.round_num != self.total_rounds:
            if self.aggregated_model_accuracy > self.top_model_accuracy:
                print(
                    (
                        "Accuracy improved to "
                        f"{self.aggregated_model_accuracy} for round {self.round_num}"
                    )
                )
                self.top_model_accuracy = self.aggregated_model_accuracy
            self.round_num += 1
            print(20 * "#")
            print(f"Round {self.round_num}...")
            print(20 * "#")
            self.next(
                self.aggregated_model_validation,
                foreach="collaborators",
                exclude=["private"],
            )
        else:
            self.next(self.end)

    @aggregator
    def end(self):
        print(20 * "#")
        print("All rounds completed successfully")
        print(20 * "#")
        print("This is the end of the flow")
        print(20 * "#")


if __name__ == "__main__":
    argparser = argparse.ArgumentParser(description=__doc__)
    argparser.add_argument(
        "--audit_dataset_ratio",
        type=float,
        default=0.2,
        help="Indicate the what fraction of the sample will be used for auditing",
    )
    argparser.add_argument(
        "--test_dataset_ratio",
        type=float,
        default=0.4,
        help="Indicate the what fraction of the sample will be used for testing",
    )
    argparser.add_argument(
        "--train_dataset_ratio",
        type=float,
        default=0.4,
        help="Indicate the what fraction of the sample will be used for training",
    )
    argparser.add_argument(
        "--signals",
        nargs="*",
        type=str,
        default=["loss", "gradient_norm", "logits"],
        help="Indicate which signal to use for membership inference attack",
    )
    argparser.add_argument(
        "--fpr_tolerance",
        nargs="*",
        type=float,
        default=[0.1, 0.5, 0.9],
        help="Indicate false positive tolerance rates in which users are interested",
    )
    argparser.add_argument(
        "--log_dir",
        type=str,
        default="test_debug",
        help="Indicate where to save the privacy loss profile and log files during the training",
    )
    argparser.add_argument(
        "--comm_round",
        type=int,
        default=30,
        help="Indicate the communication round of FL",
    )
    argparser.add_argument(
        "--auditing_interval", type=int, default=1, help="Indicate auditing interval"
    )
    argparser.add_argument(
        "--is_features",
        type=bool,
        default=True,
        help="Indicate whether to use the gradient norm with respect to the features as a signal",
    )
    argparser.add_argument(
        "--layer_number",
        type=int,
        default=10,
        help="Indicate whether layer to compute the gradient or gradient norm",
    )
    argparser.add_argument(
        "--flow_internal_loop_test",
        type=bool,
        default=False,
        help="Indicate enabling of internal loop testing of Federated Flow",
    )
    argparser.add_argument(
        "--optimizer_type",
        type=str,
        default="SGD",
        help="Indicate optimizer to use for training",
    )

    args = argparser.parse_args()

    # Setup participants
    # If running with GPU and 1 GPU is available then
    # Set `num_gpus=0.3` to run on GPU
    aggregator = Aggregator()

    collaborator_names = ["Portland", "Seattle"]

    if torch.cuda.is_available():
        device = torch.device("cuda:0")
    else:
        device = torch.device("cpu")

    # Download and setup the train, and test dataset
    transform = transforms.Compose([transforms.ToTensor()])

    cifar_train = CIFAR10(root="./data", train=True, download=True, transform=transform)

    cifar_test = CIFAR10(root="./data", train=False, download=True, transform=transform)

    # Split the dataset in train, test, and population dataset
    N_total_samples = len(cifar_test) + len(cifar_train)
    train_dataset_size = int(N_total_samples * args.train_dataset_ratio)
    test_dataset_size = int(N_total_samples * args.test_dataset_ratio)
    audit_dataset_size = min(
        int(N_total_samples * args.audit_dataset_ratio),
        N_total_samples - test_dataset_size - train_dataset_size,
    )

    X = np.concatenate([cifar_test.data, cifar_train.data])
    Y = np.concatenate([cifar_test.targets, cifar_train.targets]).tolist()

    train_dataset = deepcopy(cifar_train)
    train_dataset.data = X[:train_dataset_size]
    train_dataset.targets = Y[:train_dataset_size]

    test_dataset = deepcopy(cifar_test)
    test_dataset.data = X[train_dataset_size: train_dataset_size + test_dataset_size]
    test_dataset.targets = Y[
        train_dataset_size: train_dataset_size + test_dataset_size
    ]

    population_dataset = deepcopy(cifar_test)
    population_dataset.data = X[-audit_dataset_size:]
    population_dataset.targets = Y[-audit_dataset_size:]

    print(
        (
            f"Dataset info (total {N_total_samples}): "
            f"train - {len(train_dataset)}, "
            f"test - {len(test_dataset)}, "
            f"audit - {len(population_dataset)}"
        )
    )

    # Split train, test, and population dataset among collaborators
    # this function will be called before executing collaborator steps
    # which will return private attributes dictionary for each collaborator
    def callable_to_initialize_collaborator_private_attributes(
        index, n_collaborators, train_ds, test_ds, population_ds, args
    ):
        # construct the training and test and population dataset
        local_train = deepcopy(train_ds)
        local_test = deepcopy(test_ds)
        local_population = deepcopy(population_ds)

        local_train.data = train_ds.data[index::n_collaborators]
        local_train.targets = train_ds.targets[index::n_collaborators]

        local_test.data = test_ds.data[index::n_collaborators]
        local_test.targets = test_ds.targets[index::n_collaborators]

        local_population.data = population_ds.data[index::n_collaborators]
        local_population.targets = population_ds.targets[index::n_collaborators]

        # initialize pm report to track the privacy loss during the training
        local_pm_info = PM_report(
            fpr_tolerance_list=args.fpr_tolerance,
            is_report_roc=True,
            level="local",
            signals=args.signals,
            log_dir=args.log_dir,
            interval=args.auditing_interval,
            other_info={
                "is_features": args.is_features,
                "layer_number": args.layer_number,
            },
        )
        global_pm_info = PM_report(
            fpr_tolerance_list=args.fpr_tolerance,
            is_report_roc=True,
            level="global",
            signals=args.signals,
            log_dir=args.log_dir,
            interval=args.auditing_interval,
            other_info={
                "is_features": args.is_features,
                "layer_number": args.layer_number,
            },
        )

        Path(local_pm_info.log_dir).mkdir(parents=True, exist_ok=True)
        Path(global_pm_info.log_dir).mkdir(parents=True, exist_ok=True)

        return {
            "local_pm_info": local_pm_info,
            "global_pm_info": global_pm_info,
            "train_dataset": local_train,
            "test_dataset": local_test,  # provide the dataset obj for the auditing purpose
            "population_dataset": local_population,
            "train_loader": torch.utils.data.DataLoader(
                local_train, batch_size=batch_size_train, shuffle=True
            ),
            "test_loader": torch.utils.data.DataLoader(
                local_test, batch_size=batch_size_test, shuffle=False
            ),
        }

    collaborators = []
    for idx, collab_name in enumerate(collaborator_names):
        collaborators.append(
            Collaborator(
                name=collab_name,
                private_attributes_callable=callable_to_initialize_collaborator_private_attributes,
                # If 1 GPU is available in the machine
                # Set `num_gpus=0.0` to `num_gpus=0.3` to run on GPU
                # with ray backend with 2 collaborators
                num_cpus=0.0,
                num_gpus=0.0,
                index=idx,
                n_collaborators=len(collaborator_names),
                train_ds=train_dataset,
                test_ds=test_dataset,
                population_ds=population_dataset,
                args=args,
            )
        )

    # Set backend='ray' to use ray-backend
    local_runtime = LocalRuntime(
        aggregator=aggregator, collaborators=collaborators, backend="single_process"
    )

    print(f"Local runtime collaborators = {local_runtime.collaborators}")

    # change to the internal flow loop
    model = Net()
    top_model_accuracy = 0
    optimizers = {
        collaborator.name: default_optimizer(model, optimizer_type=args.optimizer_type)
        for collaborator in collaborators
    }
    flflow = FederatedFlow(
        model,
        optimizers,
        device,
        args.comm_round,
        top_model_accuracy,
        args.flow_internal_loop_test,
    )

    flflow.runtime = local_runtime
    flflow.run()



# Copyright (C) 2020-2025 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

"""You may copy this file as the starting point of your own model."""

import logging

import keras

from openfl.federated import KerasTaskRunner
from src.model import CNNModel

logger = logging.getLogger(__name__)

class KerasCNN(KerasTaskRunner):
    """A basic convolutional neural network model."""

    def __init__(self, **kwargs):
        """
        Initialize.

        Args:
            **kwargs: Additional parameters to pass to the function
        """
        super().__init__(**kwargs)

        self.model = self.build_model(
            self.data_loader.get_feature_shape(),
            self.data_loader.get_num_classes(),
            **kwargs
        )

        self.initialize_tensorkeys_for_functions()

        self.model.summary(print_fn=logger.info)

    def build_model(self,
                    input_shape,
                    num_classes,
                    conv_kernel_size=(4, 4),
                    conv_strides=(2, 2),
                    conv1_channels_out=16,
                    conv2_channels_out=32,
                    final_dense_inputsize=100,
                    **kwargs):
        """
        Define the model architecture.

        Args:
            input_shape (numpy.ndarray): The shape of the data
            num_classes (int): The number of classes of the dataset

        Returns:
            keras.models.Sequential: The model defined in Keras

        """
        inputs = keras.Input(shape=input_shape)
        outputs = keras.layers.Conv2D(conv1_channels_out,
                                    kernel_size=conv_kernel_size,
                                    strides=conv_strides,
                                    activation='relu',
                                    input_shape=input_shape)(inputs)
        outputs = keras.layers.Conv2D(conv2_channels_out,
                         kernel_size=conv_kernel_size,
                         strides=conv_strides,
                         activation='relu')(outputs)

        outputs = keras.layers.Flatten()(outputs)

        outputs = keras.layers.Dense(final_dense_inputsize, activation='relu')(outputs)

        outputs = keras.layers.Dense(num_classes, activation='softmax')(outputs)

        model = CNNModel(inputs, outputs)

        model.compile(loss="categorical_crossentropy",
                      optimizer="adam",
                      metrics=["accuracy"])

        return model



# Copyright 2020-2024 Intel Corporation
# SPDX-License-Identifier: Apache-2.0


"""Task module."""

from dataclasses import dataclass, field


@dataclass
class Task:
    """Task base dataclass.

    Args:
        name (str): Name of the task.
        function_name (str): Name of the function to be executed for the task.
        task_type (str): Type of the task.
        apply_local (bool, optional): Whether to apply the task locally.
            Defaults to False.
        parameters (dict, optional): Parameters for the task. Defaults to an
            empty dictionary.
    """

    name: str
    function_name: str
    task_type: str
    apply_local: bool = False
    parameters: dict = field(default_factory=dict)  # We can expend it in the future


@dataclass
class TrainTask(Task):
    """TrainTask class.

    Args:
        name (str): Name of the task.
        function_name (str): Name of the function to be executed for the task.
        apply_local (bool, optional): Whether to apply the task locally.
            Defaults to False.
        parameters (dict, optional): Parameters for the task. Defaults to an
            empty dictionary.

    Attributes:
        task_type (str): Type of the task. Set to 'train'.
    """

    task_type: str = "train"


@dataclass
class ValidateTask(Task):
    """Validation Task class.

    Args:
        name (str): Name of the task.
        function_name (str): Name of the function to be executed for the task.
        apply_local (bool, optional): Whether to apply the task locally.
            Defaults to False.
        parameters (dict, optional): Parameters for the task. Defaults to an
            empty dictionary.

    Attributes:
        task_type (str): Type of the task. Set to 'validate'.
    """

    task_type: str = "validate"



# Copyright 2025 Intel Corporation
# SPDX-License-Identifier: Apache-2.0


import json
import os
from pathlib import Path

from openfl.federated.data.sources.azure_blob_data_source import AzureBlobDataSource
from openfl.federated.data.sources.local_data_source import LocalDataSource
from openfl.federated.data.sources.s3_data_source import S3DataSource
from openfl.federated.data.sources.verifiable_dataset_info import VerifiableDatasetInfo
from openfl.utilities.path_check import is_directory_traversal


class DataSourcesJsonParser:
    @staticmethod
    def parse(json_string: str) -> VerifiableDatasetInfo:
        """
        Parse a JSON string into a dictionary.

        Args:
            json_string (str): The JSON string to parse.

        Returns:
            VerifiableDatasetInfo: An instance of VerifiableDatasetInfo containing
            the parsed data sources.
        """
        try:
            data = json.loads(json_string)
        except json.JSONDecodeError as e:
            raise ValueError(f"Invalid JSON format: {e}")

        datasources = DataSourcesJsonParser.process_data_sources(data)
        if not datasources:
            raise ValueError("No data sources were found.")
        return VerifiableDatasetInfo(
            data_sources=datasources,
            label="",
        )

    @staticmethod
    def process_data_sources(data):
        """Process and validate data sources."""
        cwd = os.getcwd()
        datasources = []
        for source_name, source_info in data.items():
            source_type = source_info.get("type", None)
            if source_type is None:
                raise ValueError(f"Missing 'type' key in data source configuration: {source_info}")
            params = source_info.get("params", {})
            if source_type == "local":
                datasources.append(
                    DataSourcesJsonParser.process_local_source(source_name, params, cwd)
                )
            elif source_type == "s3":
                datasources.append(DataSourcesJsonParser.process_s3_source(source_name, params))
            elif source_type == "azure_blob":
                datasources.append(
                    DataSourcesJsonParser.process_azure_blob_source(source_name, params)
                )
        return [ds for ds in datasources if ds]

    @staticmethod
    def process_local_source(source_name, params, cwd):
        """Process a local data source."""
        path = params.get("path", None)
        if not path:
            raise ValueError(f"Missing 'path' parameter for local data source '{source_name}'")
        abs_path = os.path.abspath(path)
        rel_path = os.path.relpath(abs_path, cwd)
        if rel_path and not is_directory_traversal(rel_path):
            return LocalDataSource(source_name, rel_path, base_path=Path("."))
        else:
            raise ValueError(f"Invalid path for local data source '{source_name}': {path}.")

    @staticmethod
    def process_s3_source(source_name, params):
        """Process an S3 data source."""
        required_fields = ["uri", "access_key_env_name", "secret_key_env_name", "secret_name"]
        missing_fields = set(required_fields) - set(params.keys())
        if missing_fields:
            raise Exception(
                f"Missing required fields: {', '.join(missing_fields)} "
                f"for S3 data source '{source_name}'"
            )
        return S3DataSource(
            name=source_name,
            uri=params["uri"],
            endpoint=params.get("endpoint") or None,
            access_key_env_name=params["access_key_env_name"],
            secret_key_env_name=params["secret_key_env_name"],
            secret_name=params["secret_name"],
        )

    @staticmethod
    def process_azure_blob_source(source_name, params):
        """Process an Azure Blob data source."""
        required_fields = ["connection_string", "container_name"]
        missing_fields = set(required_fields) - set(params.keys())
        if missing_fields:
            raise Exception(
                f"Missing required fields: {', '.join(missing_fields)} "
                f"for Azure Blob data source '{source_name}'"
            )
        return AzureBlobDataSource(
            name=source_name,
            connection_string=params["connection_string"],
            container_name=params["container_name"],
            folder_prefix=params.get("folder_prefix") or "",
        )



# Copyright (C) 2025 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

import numpy as np
from openfl.interface.aggregation_functions.core import AggregationFunction


class AggregateHealthMetrics(AggregationFunction):
    """Aggregation logic for Smokers Health analytics."""

    def call(self, local_tensors, *_) -> dict:
        """
        Aggregates local tensors which contains mean of local health metrics such as
        heart_rate_mean, cholesterol, systolic_blood_pressure, and
        diastolic_blood_pressure which are grouped by age, sex and if they smoke or not.
        Each tensor represents local metrics for these health parameters.

        Args:
            local_tensors (list): A list of objects, each containing a `tensor` attribute
                      that represents local means for the health metrics.
            *_: Additional arguments (unused).
        Returns:
            dict: A dictionary containing the aggregated means for each health metric.
        Raises:
            ValueError: If the input list `local_tensors` is empty, indicating
            that there are no metrics to aggregate.
        """

        if not local_tensors:
            raise ValueError("No local metrics to aggregate.")

        agg_histogram = np.zeros_like(local_tensors[0].tensor)
        for local_tensor in local_tensors:
            agg_histogram += local_tensor.tensor / len(local_tensors)
        return agg_histogram



# Copyright 2020-2024 Intel Corporation
# SPDX-License-Identifier: Apache-2.0


"""openfl.experimental.workflow.utilities.stream_redirect module."""

import io
import sys
from copy import deepcopy


class RedirectStdStreamBuffer:
    """Buffer object used to store stdout and stderr.

    Attributes:
        _stdoutbuff (io.StringIO): Buffer for stdout.
        _stderrbuff (io.StringIO): Buffer for stderr.
    """

    def __init__(self):
        """Initializes the RedirectStdStreamBuffer with empty stdout and stderr
        buffers."""
        self._stdoutbuff = io.StringIO()
        self._stderrbuff = io.StringIO()

    def get_stdstream(self):
        """Returns the contents of stdout and stderr buffers.

        Returns:
            tuple: A tuple containing the contents of stdout and stderr
                buffers.
        """
        self._stdoutbuff.seek(0)
        self._stderrbuff.seek(0)

        step_stdout = deepcopy(self._stdoutbuff)
        step_stderr = deepcopy(self._stderrbuff)

        self._stdoutbuff.truncate(0)
        self._stderrbuff.truncate(0)

        return step_stdout, step_stderr


class RedirectStdStream:
    """
    Class used to intercept stdout and stderr, so that stdout and stderr is
    written to buffer as well as terminal.

    Attributes:
        __stdDestination (io.TextIOWrapper): Destination for standard outputs.
        __stdBuffer (RedirectStdStreamBuffer): Buffer for standard outputs.
    """

    def __init__(self, buffer, destination):
        """Initializes the RedirectStdStream with a buffer and a destination.

        Args:
            buffer (RedirectStdStreamBuffer): Buffer for standard outputs.
            destination (io.TextIOWrapper): Destination for standard outputs.
        """
        self.__stdDestination = destination
        self.__stdBuffer = buffer

    def write(self, message):
        """Writes the message to the standard destination and buffer.

        Args:
            message (str): The message to write.
        """
        message = f"\33[94m{message}\33[0m"
        self.__stdDestination.write(message)
        self.__stdBuffer.write(message)

    def flush(self):
        pass


class RedirectStdStreamContext:
    """Context Manager that enables redirection of stdout and stderr.

    Attributes:
        stdstreambuffer (RedirectStdStreamBuffer): Buffer for standard outputs.
    """

    def __init__(self):
        """Initializes the RedirectStdStreamContext with a
        RedirectStdStreamBuffer."""
        self.stdstreambuffer = RedirectStdStreamBuffer()

    def __enter__(self):
        """Creates a context to redirect stdout and stderr.

        Returns:
            RedirectStdStreamBuffer: The buffer for standard outputs.
        """
        self.__old_stdout = sys.stdout
        self.__old_stderr = sys.stderr
        sys.stdout = RedirectStdStream(self.stdstreambuffer._stdoutbuff, sys.stdout)
        sys.stderr = RedirectStdStream(self.stdstreambuffer._stderrbuff, sys.stderr)

        return self.stdstreambuffer

    def __exit__(self, et, ev, tb):
        """Exits the context and restores the stdout and stderr.

        Args:
            et (type): The type of exception.
            ev (BaseException): The instance of exception.
            tb (traceback): A traceback object encapsulating the call stack.
        """
        sys.stdout = self.__old_stdout
        sys.stderr = self.__old_stderr



from openfl.protocols import base_pb2 as _base_pb2
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class MessageHeader(_message.Message):
    __slots__ = ("sender", "receiver", "federation_uuid", "single_col_cert_common_name")
    SENDER_FIELD_NUMBER: _ClassVar[int]
    RECEIVER_FIELD_NUMBER: _ClassVar[int]
    FEDERATION_UUID_FIELD_NUMBER: _ClassVar[int]
    SINGLE_COL_CERT_COMMON_NAME_FIELD_NUMBER: _ClassVar[int]
    sender: str
    receiver: str
    federation_uuid: str
    single_col_cert_common_name: str
    def __init__(self, sender: _Optional[str] = ..., receiver: _Optional[str] = ..., federation_uuid: _Optional[str] = ..., single_col_cert_common_name: _Optional[str] = ...) -> None: ...

class TaskResultsRequest(_message.Message):
    __slots__ = ("header", "collab_name", "round_number", "next_step", "execution_environment")
    HEADER_FIELD_NUMBER: _ClassVar[int]
    COLLAB_NAME_FIELD_NUMBER: _ClassVar[int]
    ROUND_NUMBER_FIELD_NUMBER: _ClassVar[int]
    NEXT_STEP_FIELD_NUMBER: _ClassVar[int]
    EXECUTION_ENVIRONMENT_FIELD_NUMBER: _ClassVar[int]
    header: MessageHeader
    collab_name: str
    round_number: int
    next_step: str
    execution_environment: bytes
    def __init__(self, header: _Optional[_Union[MessageHeader, _Mapping]] = ..., collab_name: _Optional[str] = ..., round_number: _Optional[int] = ..., next_step: _Optional[str] = ..., execution_environment: _Optional[bytes] = ...) -> None: ...

class TaskResultsResponse(_message.Message):
    __slots__ = ("header",)
    HEADER_FIELD_NUMBER: _ClassVar[int]
    header: MessageHeader
    def __init__(self, header: _Optional[_Union[MessageHeader, _Mapping]] = ...) -> None: ...

class GetTasksRequest(_message.Message):
    __slots__ = ("header",)
    HEADER_FIELD_NUMBER: _ClassVar[int]
    header: MessageHeader
    def __init__(self, header: _Optional[_Union[MessageHeader, _Mapping]] = ...) -> None: ...

class GetTasksResponse(_message.Message):
    __slots__ = ("header", "round_number", "function_name", "execution_environment", "sleep_time", "quit")
    HEADER_FIELD_NUMBER: _ClassVar[int]
    ROUND_NUMBER_FIELD_NUMBER: _ClassVar[int]
    FUNCTION_NAME_FIELD_NUMBER: _ClassVar[int]
    EXECUTION_ENVIRONMENT_FIELD_NUMBER: _ClassVar[int]
    SLEEP_TIME_FIELD_NUMBER: _ClassVar[int]
    QUIT_FIELD_NUMBER: _ClassVar[int]
    header: MessageHeader
    round_number: int
    function_name: str
    execution_environment: bytes
    sleep_time: int
    quit: bool
    def __init__(self, header: _Optional[_Union[MessageHeader, _Mapping]] = ..., round_number: _Optional[int] = ..., function_name: _Optional[str] = ..., execution_environment: _Optional[bytes] = ..., sleep_time: _Optional[int] = ..., quit: bool = ...) -> None: ...

class CheckpointRequest(_message.Message):
    __slots__ = ("header", "execution_environment", "function", "stream_buffer")
    HEADER_FIELD_NUMBER: _ClassVar[int]
    EXECUTION_ENVIRONMENT_FIELD_NUMBER: _ClassVar[int]
    FUNCTION_FIELD_NUMBER: _ClassVar[int]
    STREAM_BUFFER_FIELD_NUMBER: _ClassVar[int]
    header: MessageHeader
    execution_environment: bytes
    function: bytes
    stream_buffer: bytes
    def __init__(self, header: _Optional[_Union[MessageHeader, _Mapping]] = ..., execution_environment: _Optional[bytes] = ..., function: _Optional[bytes] = ..., stream_buffer: _Optional[bytes] = ...) -> None: ...

class CheckpointResponse(_message.Message):
    __slots__ = ("header",)
    HEADER_FIELD_NUMBER: _ClassVar[int]
    header: MessageHeader
    def __init__(self, header: _Optional[_Union[MessageHeader, _Mapping]] = ...) -> None: ...



from openfl.protocols import base_pb2 as _base_pb2
from google.protobuf import timestamp_pb2 as _timestamp_pb2
from google.protobuf import duration_pb2 as _duration_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Iterable as _Iterable, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class SendConnectionRequest(_message.Message):
    __slots__ = ("envoy_name",)
    ENVOY_NAME_FIELD_NUMBER: _ClassVar[int]
    envoy_name: str
    def __init__(self, envoy_name: _Optional[str] = ...) -> None: ...

class RequestAccepted(_message.Message):
    __slots__ = ("accepted",)
    ACCEPTED_FIELD_NUMBER: _ClassVar[int]
    accepted: bool
    def __init__(self, accepted: bool = ...) -> None: ...

class WaitExperimentRequest(_message.Message):
    __slots__ = ("collaborator_name",)
    COLLABORATOR_NAME_FIELD_NUMBER: _ClassVar[int]
    collaborator_name: str
    def __init__(self, collaborator_name: _Optional[str] = ...) -> None: ...

class WaitExperimentResponse(_message.Message):
    __slots__ = ("experiment_name",)
    EXPERIMENT_NAME_FIELD_NUMBER: _ClassVar[int]
    experiment_name: str
    def __init__(self, experiment_name: _Optional[str] = ...) -> None: ...

class GetExperimentDataRequest(_message.Message):
    __slots__ = ("experiment_name", "collaborator_name")
    EXPERIMENT_NAME_FIELD_NUMBER: _ClassVar[int]
    COLLABORATOR_NAME_FIELD_NUMBER: _ClassVar[int]
    experiment_name: str
    collaborator_name: str
    def __init__(self, experiment_name: _Optional[str] = ..., collaborator_name: _Optional[str] = ...) -> None: ...

class ExperimentData(_message.Message):
    __slots__ = ("size", "exp_data")
    SIZE_FIELD_NUMBER: _ClassVar[int]
    EXP_DATA_FIELD_NUMBER: _ClassVar[int]
    size: int
    exp_data: bytes
    def __init__(self, size: _Optional[int] = ..., exp_data: _Optional[bytes] = ...) -> None: ...

class UpdateEnvoyStatusRequest(_message.Message):
    __slots__ = ("name", "is_experiment_running")
    NAME_FIELD_NUMBER: _ClassVar[int]
    IS_EXPERIMENT_RUNNING_FIELD_NUMBER: _ClassVar[int]
    name: str
    is_experiment_running: bool
    def __init__(self, name: _Optional[str] = ..., is_experiment_running: bool = ...) -> None: ...

class UpdateEnvoyStatusResponse(_message.Message):
    __slots__ = ("health_check_period",)
    HEALTH_CHECK_PERIOD_FIELD_NUMBER: _ClassVar[int]
    health_check_period: _duration_pb2.Duration
    def __init__(self, health_check_period: _Optional[_Union[_duration_pb2.Duration, _Mapping]] = ...) -> None: ...

class ExperimentInfo(_message.Message):
    __slots__ = ("name", "collaborator_names", "experiment_data")
    NAME_FIELD_NUMBER: _ClassVar[int]
    COLLABORATOR_NAMES_FIELD_NUMBER: _ClassVar[int]
    EXPERIMENT_DATA_FIELD_NUMBER: _ClassVar[int]
    name: str
    collaborator_names: _containers.RepeatedScalarFieldContainer[str]
    experiment_data: ExperimentData
    def __init__(self, name: _Optional[str] = ..., collaborator_names: _Optional[_Iterable[str]] = ..., experiment_data: _Optional[_Union[ExperimentData, _Mapping]] = ...) -> None: ...

class SetNewExperimentResponse(_message.Message):
    __slots__ = ("status",)
    STATUS_FIELD_NUMBER: _ClassVar[int]
    status: bool
    def __init__(self, status: bool = ...) -> None: ...

class EnvoyInfo(_message.Message):
    __slots__ = ("envoy_name", "experiment_name", "is_online", "is_experiment_running", "last_updated", "valid_duration")
    ENVOY_NAME_FIELD_NUMBER: _ClassVar[int]
    EXPERIMENT_NAME_FIELD_NUMBER: _ClassVar[int]
    IS_ONLINE_FIELD_NUMBER: _ClassVar[int]
    IS_EXPERIMENT_RUNNING_FIELD_NUMBER: _ClassVar[int]
    LAST_UPDATED_FIELD_NUMBER: _ClassVar[int]
    VALID_DURATION_FIELD_NUMBER: _ClassVar[int]
    envoy_name: str
    experiment_name: str
    is_online: bool
    is_experiment_running: bool
    last_updated: _timestamp_pb2.Timestamp
    valid_duration: _duration_pb2.Duration
    def __init__(self, envoy_name: _Optional[str] = ..., experiment_name: _Optional[str] = ..., is_online: bool = ..., is_experiment_running: bool = ..., last_updated: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ..., valid_duration: _Optional[_Union[_duration_pb2.Duration, _Mapping]] = ...) -> None: ...

class GetEnvoysRequest(_message.Message):
    __slots__ = ()
    def __init__(self) -> None: ...

class GetEnvoysResponse(_message.Message):
    __slots__ = ("envoy_infos",)
    ENVOY_INFOS_FIELD_NUMBER: _ClassVar[int]
    envoy_infos: _containers.RepeatedCompositeFieldContainer[EnvoyInfo]
    def __init__(self, envoy_infos: _Optional[_Iterable[_Union[EnvoyInfo, _Mapping]]] = ...) -> None: ...

class GetFlowStateRequest(_message.Message):
    __slots__ = ()
    def __init__(self) -> None: ...

class GetFlowStateResponse(_message.Message):
    __slots__ = ("completed", "flspec_obj")
    COMPLETED_FIELD_NUMBER: _ClassVar[int]
    FLSPEC_OBJ_FIELD_NUMBER: _ClassVar[int]
    completed: bool
    flspec_obj: bytes
    def __init__(self, completed: bool = ..., flspec_obj: _Optional[bytes] = ...) -> None: ...

class SendRuntimeRequest(_message.Message):
    __slots__ = ()
    def __init__(self) -> None: ...

class RuntimeRequestResponse(_message.Message):
    __slots__ = ("accepted",)
    ACCEPTED_FIELD_NUMBER: _ClassVar[int]
    accepted: bool
    def __init__(self, accepted: bool = ...) -> None: ...

class GetExperimentStdoutRequest(_message.Message):
    __slots__ = ("experiment_name",)
    EXPERIMENT_NAME_FIELD_NUMBER: _ClassVar[int]
    experiment_name: str
    def __init__(self, experiment_name: _Optional[str] = ...) -> None: ...

class GetExperimentStdoutResponse(_message.Message):
    __slots__ = ("stdout_origin", "task_name", "stdout_value")
    STDOUT_ORIGIN_FIELD_NUMBER: _ClassVar[int]
    TASK_NAME_FIELD_NUMBER: _ClassVar[int]
    STDOUT_VALUE_FIELD_NUMBER: _ClassVar[int]
    stdout_origin: str
    task_name: str
    stdout_value: str
    def __init__(self, stdout_origin: _Optional[str] = ..., task_name: _Optional[str] = ..., stdout_value: _Optional[str] = ...) -> None: ...



# Copyright 2020-2024 Intel Corporation
# SPDX-License-Identifier: Apache-2.0


import os
import webbrowser
from pathlib import Path

from openfl.experimental.workflow.utilities.metaflow_utils import DefaultCard, FlowGraph


class InspectFlow:
    """Class for inspecting a flow.

    Attributes:
        ds_root (str): The root directory for the data store. Defaults to
            "~/.metaflow".
        show_html (bool): Whether to show the UI in a web browser. Defaults to
            False.
        run_id (str): The run ID of the flow.
        flow_name (str): The name of the flow.
        graph_dict (dict): The graph of the flow.
    """

    def __init__(
        self,
        flow_obj,
        run_id,
        show_html=False,
        ds_root=None,
    ):
        """Initializes the InspectFlow with a flow object, run ID, an optional
        flag to show the UI in a web browser, and an optional root directory
        for the data store.

        Args:
            flow_obj (Flow): The flow object to inspect.
            run_id (str): The run ID of the flow.
            show_html (bool, optional): Whether to show the UI in a web
                browser. Defaults to False.
            ds_root (str, optional): The root directory for the data store.
                Defaults to "~/.metaflow".
        """
        self.ds_root = ds_root or f"{Path.home()}/.metaflow"
        self.show_html = show_html
        self.run_id = run_id
        self.flow_name = flow_obj.__class__.__name__
        self._graph = FlowGraph(flow_obj.__class__)
        self._steps = [getattr(flow_obj, node.name) for node in self._graph]

        self.graph_dict, _ = self._graph.output_steps()
        self.show_ui()

    def get_pathspec(self):
        """Gets the path specification of the flow.

        Returns:
            str: The path specification of the flow.
        """
        return f"{self.ds_root}/{self.flow_name}/{self.run_id}"

    def open_in_browser(self, card_path):
        """Opens the specified path in a web browser.

        Args:
            card_path (str): The path to open.
        """
        url = "file://" + os.path.abspath(card_path)
        webbrowser.open(url)

    def show_ui(self):
        """Shows the UI of the flow in a web browser if show_html is True, and
        saves the UI as an HTML file."""

        default_card = DefaultCard(graph=self.graph_dict)

        pathspec = self.get_pathspec()
        print(f"Flowgraph generated at :{pathspec}")
        html = default_card.render(pathspec)

        with open(f"{pathspec}/card_ui.html", "w") as f:
            f.write(html)

        if self.show_html:
            self.open_in_browser(f"{pathspec}/card_ui.html")



# Copyright 2020-2024 Intel Corporation
# SPDX-License-Identifier: Apache-2.0


"""split tensors module."""

import logging

import numpy as np

logger = logging.getLogger(__name__)


def split_tensor_dict_into_floats_and_non_floats(tensor_dict):
    """Split the tensor dictionary into float and non-floating point values.

    This function splits a tensor dictionary into two dictionaries: one
    containing all the floating point tensors and the other containing all the
    non-floating point tensors.

    Args:
        tensor_dict (dict): A dictionary of tensors.

    Returns:
        Tuple[dict, dict]: The first dictionary contains all of the floating
            point tensors and the second dictionary contains all of the
            non-floating point tensors.
    """
    float_dict = {}
    non_float_dict = {}
    for k, v in tensor_dict.items():
        if np.issubdtype(v.dtype, np.floating):
            float_dict[k] = v
        else:
            non_float_dict[k] = v
    return float_dict, non_float_dict


def split_tensor_dict_by_types(tensor_dict, keep_types):
    """Split the tensor dictionary into supported and not supported types.

    Args:
        tensor_dict (dict): A dictionary of tensors.
        keep_types (Iterable[type]): An iterable of supported types.

    Returns:
        Tuple[dict, dict]: The first dictionary contains all of the supported
            tensors and the second dictionary contains all of the not
            supported tensors.
    """
    keep_dict = {}
    holdout_dict = {}
    for k, v in tensor_dict.items():
        if any(np.issubdtype(v.dtype, type_) for type_ in keep_types):
            keep_dict[k] = v
        else:
            holdout_dict[k] = v
    return keep_dict, holdout_dict


def split_tensor_dict_for_holdouts(
    tensor_dict,
    keep_types=(np.floating, np.integer),
    holdout_tensor_names=(),
):
    """
    Split a tensor according to tensor types.

    This function splits a tensor dictionary into two dictionaries: one
    containing the tensors to send and the other containing the holdout
    tensors.

    Args:
        tensor_dict (dict): A dictionary of tensors.
        keep_types (Tuple[type, ...], optional): A tuple of types to keep in
            the dictionary of tensors. Defaults to (np.floating, np.integer).
        holdout_tensor_names (Iterable[str], optional): An iterable of tensor
            names to extract from the dictionary of tensors. Defaults to ().

    Returns:
        Tuple[dict, dict]: The first dictionary is the original tensor
            dictionary minus the holdout tensors and the second dictionary is
            a tensor dictionary with only the holdout tensors.
    """
    # initialization
    tensors_to_send = tensor_dict.copy()
    holdout_tensors = {}

    # filter by-name tensors from tensors_to_send and add to holdout_tensors
    # (for ones not already held out because of their type)
    for tensor_name in holdout_tensor_names:
        if tensor_name not in holdout_tensors.keys():
            try:
                holdout_tensors[tensor_name] = tensors_to_send.pop(tensor_name)
            except KeyError:
                logger.warning(
                    f"tried to remove tensor: {tensor_name} not present in the tensor dict"
                )
                continue

    # filter holdout_types from tensors_to_send and add to holdout_tensors
    tensors_to_send, not_supported_tensors_dict = split_tensor_dict_by_types(
        tensors_to_send, keep_types
    )
    holdout_tensors = {**holdout_tensors, **not_supported_tensors_dict}

    return tensors_to_send, holdout_tensors



# Copyright 2020-2024 Intel Corporation
# SPDX-License-Identifier: Apache-2.0


"""Logs utilities."""

import logging

from rich.console import Console
from rich.logging import RichHandler

from openfl.utilities import add_log_level


def setup_logger(log_level=logging.INFO, log_file=None):
    """Configure loggers.

    This function sets up the root logger to log messages with a certain
    minimum level and a specific format.

    Args:
        log_level (int, optional): The minimum level of messages to log.
            Defaults to logging.INFO.
        log_file (str, optional): The file to which log messages should be written.
    """
    # Create a logger instance
    logger = logging.getLogger()

    # Add a custom log level for METRIC
    metric = 25
    add_log_level("METRIC", metric)

    if isinstance(log_level, str):
        log_level = log_level.upper()

    # Set the log level for the logger
    logger.setLevel(log_level)

    console = Console(width=160)
    console_handler = RichHandler(console=console)

    # Console handler includes date and log level, do not add it again
    console_handler.setFormatter(logging.Formatter("%(message)s"))
    logger.addHandler(console_handler)

    # Create a file handler if log_file is provided
    if log_file:
        file_formatter = logging.Formatter(
            "[%(asctime)s] [%(filename)s:%(lineno)d] %(levelname)s %(message)s",
            datefmt="%H:%M:%S",
        )
        file_handler = logging.FileHandler(log_file)
        file_handler.setFormatter(file_formatter)
        logger.addHandler(file_handler)

    return logger



# Copyright 2025 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

"""This module contains the VerifiableDatasetInfo class."""

import json
from hashlib import sha384
from typing import List

from openfl.federated.data.sources.azure_blob_data_source import AzureBlobDataSource
from openfl.federated.data.sources.data_source import DataSource, DataSourceType
from openfl.federated.data.sources.local_data_source import LocalDataSource
from openfl.federated.data.sources.s3_data_source import S3DataSource


class VerifiableDatasetInfo:
    """
    This class represents a data set whose integrity can be verified.
    It contains multiple data sources and methods to compute data commitments, verification,
    as well as utilities for serialization and deserialization.
    """

    def __init__(
        self,
        data_sources: List[DataSource],
        label: str,
        metadata=None,
        root_hash=None,
    ):
        self.data_sources = data_sources
        self.label = label
        self.metadata = metadata
        self.root_hash = root_hash
        self.all_hashes = None

    def _create_verbose_dataset_hash(self):
        all_hashes = {
            str(file_path): ds.compute_file_hash(str(file_path))
            for ds in self.data_sources
            for file_path in ds.enumerate_files()
        }
        return all_hashes

    def _create_concise_dataset_hash(self):
        all_file_hashes = self._create_verbose_dataset_hash()
        sorted_file_hashes = sorted(all_file_hashes.values())
        joined_hashes = "".join(sorted_file_hashes)
        root_hash = sha384(joined_hashes.encode()).hexdigest()
        return root_hash, all_file_hashes

    def create_dataset_hash(self):
        """Create and return the root_hash of all files hashes."""
        root_hash, _ = self._create_concise_dataset_hash()
        return root_hash

    def _validate_verbose_dataset_info(self):
        hashes = self._create_verbose_dataset_hash()
        return sorted(hashes.values()) == sorted(self.all_hashes.values())

    def _validate_concise_dataset_info(self):
        concise_hash, all_hashes = self._create_concise_dataset_hash()
        valid = concise_hash == self.root_hash
        if valid and self.all_hashes is None:
            self.all_hashes = all_hashes
        return valid

    def verify_dataset(self, root_hash=None):
        """Verify the dataset against root_hash."""
        if root_hash is None and self.root_hash is None:
            raise ValueError("No saved root hash found. Please provide 'dataset_info'.")
        if root_hash:
            self.root_hash = root_hash
        return self._validate_concise_dataset_info()

    def _verify_file_verbose(self, file_path, file_hash):
        if str(file_path) not in self.all_hashes:
            raise KeyError(f"Verification failed: No information found for the file: {file_path}")
        return self.all_hashes[str(file_path)] == file_hash

    def verify_single_file(self, file_path, file_hash):
        """Verify the hash of a single file."""
        if self.all_hashes is None:
            if self.root_hash is not None:
                self.verify_dataset(self.root_hash)
            else:
                raise ValueError("Trusted hash not found in the dataset")
        return self._verify_file_verbose(file_path, file_hash)

    def to_json(self):
        if len(self.data_sources) == 1:
            return self._to_json_v1()
        return self._to_json_v2()

    def _to_json_v1(self):
        if self.data_sources[0].type == DataSourceType.LOCAL:
            path = str(self.data_sources[0].get_source_full_path())
        elif self.data_sources[0].type == DataSourceType.S3:
            path = self.data_sources[0].uri
        elif self.data_sources[0].type == DataSourceType.AZURE_BLOB:
            path = self.data_sources[0].connection_string
        else:
            raise ValueError(f"Unknown storage type: {self.data_sources[0].type}")

        dataset_dict = {
            "dataset_id": self.create_dataset_hash(),
            "mount_absolute_path": path,
            "label": self.label,
            "metadata": self.metadata,
            "dataset_format": "concise_dataset",
            "name": self.data_sources[0].name,
        }
        if self.data_sources[0].type == DataSourceType.S3:
            s3_dict = self.data_sources[0].to_dict()
            s3_dict.pop("uri")
            dataset_dict.update(s3_dict)
        elif self.data_sources[0].type == DataSourceType.AZURE_BLOB:
            azure_dict = self.data_sources[0].to_dict()
            azure_dict.pop("connection_string")
            dataset_dict.update(azure_dict)

        return json.dumps(dataset_dict, sort_keys=True, indent=4)

    def _to_json_v2(self):
        dataset_dict = {
            "version": "2.0",
            "data_sources": [ds.to_dict() for ds in self.data_sources],
            "label": self.label,
            "metadata": self.metadata,
        }
        dataset_dict["root_hash"] = self.create_dataset_hash()
        return json.dumps(dataset_dict, sort_keys=True, indent=4)

    def _from_dict_v1(data_dict, base_path=None):
        if "endpoint" in data_dict:
            s3_dict = data_dict
            s3_dict["uri"] = data_dict["mount_absolute_path"]
            return VerifiableDatasetInfo(
                [S3DataSource.from_dict(ds_dict=s3_dict)],
                label=data_dict["label"],
                metadata=data_dict["metadata"],
                root_hash=data_dict["dataset_id"],
            )
        elif "container_name" in data_dict:
            azure_dict = data_dict
            azure_dict["connection_string"] = data_dict["mount_absolute_path"]
            return VerifiableDatasetInfo(
                [AzureBlobDataSource.from_dict(ds_dict=azure_dict)],
                label=data_dict["label"],
                metadata=data_dict["metadata"],
                root_hash=data_dict["dataset_id"],
            )
        else:
            return VerifiableDatasetInfo(
                [
                    LocalDataSource(
                        name=data_dict.get("name", None), source_path=".", base_path=base_path
                    )
                ],
                label=data_dict["label"],
                metadata=data_dict["metadata"],
                root_hash=data_dict["dataset_id"],
            )

    @staticmethod
    def _from_dict_v2(data_dict, base_path=None):
        """Deserialize the VerifiableDatasetInfo from JSON"""

        # Create appropriate data source based on dictionary information
        data_sources = []
        for datasource in data_dict["data_sources"]:
            if datasource["type"] == DataSourceType.LOCAL.value:
                data_source = LocalDataSource.from_dict(ds_dict=datasource, base_path=base_path)
            elif datasource["type"] == DataSourceType.S3.value:
                data_source = S3DataSource.from_dict(ds_dict=datasource)
            elif datasource["type"] == DataSourceType.AZURE_BLOB.value:
                data_source = AzureBlobDataSource.from_dict(ds_dict=datasource)
            else:
                raise ValueError(f"Unknown storage type: {datasource['type']}")
            data_sources.append(data_source)

        return VerifiableDatasetInfo(
            data_sources,
            label=data_dict["label"],
            metadata=data_dict["metadata"],
            root_hash=data_dict["root_hash"],
        )

    @staticmethod
    def _from_dict(data_dict, base_path=None):
        """Deserialize the VerifiableDatasetInfo from JSON"""
        if "version" in data_dict and data_dict["version"] == "2.0":
            return VerifiableDatasetInfo._from_dict_v2(data_dict, base_path)
        return VerifiableDatasetInfo._from_dict_v1(data_dict, base_path)

    @staticmethod
    def from_json(json_str, base_path=None):
        """Deserialize the VerifiableDatasetInfo from JSON"""
        data_dict = json.loads(json_str)
        return VerifiableDatasetInfo._from_dict(data_dict, base_path)



# Copyright 2020-2024 Intel Corporation
# SPDX-License-Identifier: Apache-2.0


from openfl.utilities.optimizers.numpy.adagrad_optimizer import NumPyAdagrad
from openfl.utilities.optimizers.numpy.adam_optimizer import NumPyAdam
from openfl.utilities.optimizers.numpy.yogi_optimizer import NumPyYogi



# Copyright 2020-2024 Intel Corporation
# SPDX-License-Identifier: Apache-2.0


"""PyTorch optimizers package."""

from importlib import util

if util.find_spec("torch") is not None:
    from openfl.utilities.optimizers.torch.fedprox import FedProxAdam  # NOQA
    from openfl.utilities.optimizers.torch.fedprox import FedProxOptimizer  # NOQA
